# -*- coding: utf-8 -*-
#
# Simulation class
# Copyright (c) 2019-2022 Klaus K. Holst.  All rights reserved.

import kvaser as kv
import networkx as nx
import numpy as np
import pandas as pd
from PIL import Image
from io import BytesIO
import pydot

class dag:
    r"""DAG model class
    """

    def __init__(self):
        r"""
        """
        self.G = nx.DiGraph()
        self.var = {}

    def regression(self, y, x=[]):
        self.G.add_node(y)
        if y not in self.var.keys():
            self.distribution(y)
        for v in x:
            self.G.add_edge(v, y)
            if v not in self.var.keys():
                self.distribution(v)

    def distribution(self, y, generator=kv.normal()):
        r"""Set distribution of variable"""
        self.G.add_node(y)
        self.var[y] = generator

    def simulate(self, n=1, p={}):
        deg = dict(self.G.in_degree)
        vv = list(self.G.nodes)
        res = np.ndarray((n, len(vv)))

        p0 = {}
        for y in vv:
            par = self.G.predecessors(y)
            for x in par:
                pname = y + '~' + x
                p0[pname] =1
                if pname in p.keys():
                    p0[pname] = p[pname]

        while any(x>=0 for x in deg.values()):
            for v, d in deg.items():
                if d>=0:
                    par = list(self.G.predecessors(v))
                    subdict = dict((k, deg[k]) for k in par if k in deg)
                    if all(x<0 for x in subdict.values()):
                        deg[v] = -1
                        pos = vv.index(v)
                        lp = np.repeat([0.0], n)
                        # print("Variable :" + v)
                        for x in par:
                            pname = v + '~' + x
                            posx = int(vv.index(x))
                            lp += p0[pname]*res[:,posx]
                        y = np.float64(self.var[v].simulate(lp=lp))
                        res[:,pos] = y
        df = pd.DataFrame(res)
        df.columns = vv
        return df

    def plot(self):
        pdot = nx.nx_pydot.to_pydot(self.G)
        Image.open(BytesIO(pdot.create_png())).show()

    def __str__(self):
        st = ''
        for v in self.G.nodes:
            parents = list(self.G.predecessors(v))
            st += v + ' ~ ' + ' + '.join(parents) + '\n'
        st += '\n'
        for k,v in self.var.items():
            st += str(k) + ': ' + str(v) + '\n'
        return st


# m = dag()
# m.regression('y', ['x','z'])
# m.regression('z', ['x'])
# m.distribution('y', normal(scale=2))
# m.distribution('z', poisson())
# m.distribution('x', bernoulli())
# print(m)
# r = m.simulate(5000)
# r

# import statsmodels.formula.api as smf
# import statsmodels.genmod.families as fam
# smf.glm('y ~ x+z', data=r, family=fam.Gaussian()).fit().summary()

# smf.glm('z ~ x', data=r, family=fam.Poisson()).fit().summary()

# smf.glm('x ~ 1', data=r, family=fam.Binomial()).fit().summary()



# -*- coding: utf-8 -*-
#
# Time series utilities
# Copyright (c) 2019-2022 Klaus K. Holst.  All rights reserved.

import datetime
import pytz
import math
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go


def now(tz='Europe/Copenhagen'):
    if tz is None:
        for i in pytz.common_timezones:
            print(i)
        return None
    return pd.to_datetime(datetime.datetime.now()).tz_localize(tz)


def ts(x, index=None, tz='Europe/Copenhagen'):
    if not isinstance(x, (pd.Series, pd.DataFrame)):
        try:
            x = pd.DataFrame(x)
        except Exception as ex:
            raise TypeError("Pandas type expected") from ex
    if index is None:
        index = x.index
    else:
        index = pd.to_datetime(index)
        if index.dt.tz is None:
            index = index.dt.tz_localize(tz)
        else:
            index = index.dt.tz_convert(tz)
        x.index = index
    return x


def plot(y, upper=None, lower=None,
         interval_text='Prediction interval',
         highlight_start=None, highlight_text="",
         highlight_color="MediumPurple", highlight_fill="Gray",
         line_shape="hvh",
         select_fraction=0.2,
         title=None):
    r"""Plot time series

    See Also
    ----------
    plotly.express.scatter

    Returns
    ----------
    plotly.graph_objs._figure.Figure
        plotly object

    Parameters
    ----------
    y: pandas.Series
        Series or DataFrame with time series in each column
    upper: pandas.Series
        Upper prediction limit (single series)
    lower: pandas.Series
        Lower prediction limit (single series)
    interval_text: str
    highlight_start: str
    highlight_text: str
    highlight_color: str
    highlight_fill: str
    line_shape: str
    select_fraction: float
    title: str
    """

    if not isinstance(y, (pd.Series, pd.DataFrame)):
        try:
            y = pd.DataFrame(y)
        except Exception as ex:
            raise TypeError("Pandas type expected") from ex
    time = y.index
    first = min(time)
    last = max(time)
    if isinstance(y, pd.Series):
        fig = px.line(x=time, y=y,
                      title=title)
    else:
        fig = px.line(y, x=time, y=y.columns,
                      render_mode='auto',
                      title=title)
    if upper is not None and lower is not None:
        fig.add_traces([go.Scatter(x=time, y=upper,
                                   mode='lines', line_color='rgba(0,0,0,0)',
                                   showlegend=False),
                        go.Scatter(x=time, y=lower,
                                   mode='lines', line_color='rgba(0,0,0,0)',
                                   name=interval_text,
                                   fill='tonexty', fillcolor='rgba(0, 0, 255, 0.2)')])

    if highlight_start is not None:
        fig.update_layout(shapes=[dict(type='line',
                                       yref='paper', y0=0, y1=1,
                                       xref='x', x0=highlight_start, x1=highlight_start,
                                       line=dict(color=highlight_color,
                                                 width=3,
                                                 dash="dot")
                                    )])
        fig.add_vrect(x0=first, x1=highlight_start,
                      fillcolor=highlight_fill, opacity=0.2, line_width=0)
        fig.add_vrect(x0=highlight_start, x1=last,
                      annotation_text=highlight_text, annotation_position="top left",
                      annotation_font_size=11,
                      annotation_font_color=highlight_color,
                      line_width=0)
    fig.update_traces(line_shape=line_shape)
    select_visible = False
    if select_fraction > 0:
        sel_index = math.floor(len(time)*(1-select_fraction))
        select_visible = True
        fig.update_xaxes(
            range=[time[sel_index], last],
            rangeslider=dict(visible=True,
                             range=[first, last]),
            rangeselector={
                'visible': select_visible,
            })
    return fig



#!/usr/bin/env python3

import numpy as np

class Dist:
    r"""Probability distribution class.
    Super class, not to be called directly
    """

    rng = np.random.default_rng()
    name = 'Generic'
    meanpar = 'loc'

    def gen(self, size, **kwargs):
        return None

    @staticmethod
    def invlink(x):
        return np.array(x)

    def __init__(self, seed=None, **kwargs):
        r"""Distribution class constructor

        Parameters
        ----------
        seed: int64
              Random seed (optional)
        **kwargs:
              Extra arguments passed to random generator

        Returns
        ----------
        Dist
           Dist object
        """
        self.kparam = kwargs
        self.rng = np.random.default_rng(seed)

    def simulate(self, mean=None, lp=None):
        r"""Simulation method

        Parameters
        ----------
        mean:

        Returns
        ----------

        """
        if lp is not None:
            mean = self.invlink(lp)
        par = {}
        par[self.meanpar] = mean

        return self.gen(**par, **self.kparam, size=len(mean))

    def __repr__(self):
        return self.name + ' distribution ' + \
            str(self.kparam)

    def __str__(self):
        return self.name + ' distribution ' + \
            str(self.kparam)


class normal(Dist):
    r"""Normal distribution class
    constructor arguments:
    loc: float or array_like of floats
         Mean
    scale: float or array_like of floats
         Standard deviation
    """
    def gen(self, size, **kwargs):
        return self.rng.normal(size=size, **kwargs)
    name = 'Normal'
    meanpar = 'loc'


class bernoulli(Dist):
    name = 'Binomial'
    meanpar = 'p'

    def gen(self, size, **kwargs):
        return self.rng.binomial(size=size, n=1, **kwargs)

    @staticmethod
    def invlink(x):
        return(1/(1+np.exp(-np.array(x))))


class poisson(Dist):
    name = 'Poisson'
    meanpar = 'lam'
    def gen(self, size, **kwargs):
        return self.rng.poisson(size=size, **kwargs)

    @staticmethod
    def invlink(x):
        return(np.exp(np.array(x)))


class discrete(Dist):
    name = 'Discrete'

    def __init__(self, values=[0,1], p=[0.5,0.5], **kwargs):
        super().__init__()
        self.values = np.array(values)
        self.p = np.array(p)

    def gen(self, size, **kwargs):
        return self.rng.choice(a=self.values, p=self.p, size=size)



